/* lalonde
 * Copyright (C) 2010 Noé Falzon <firstname.lastname@aliceadsl.fr>
 *
 * This program is  free software:  you can redistribute it and/or modify it
 * under  the terms  of the  GNU General Public License  as published by the
 * Free Software Foundation, version 3 of the License.
 *
 * This program  is  distributed  in the hope  that it will  be useful,  but
 * WITHOUT  ANY WARRANTY;  without  even the implied  warranty of MERCHANTA-
 * BILITY  or  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the  GNU General Public License  along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "SDL.h"
#include "SDL_opengl.h"
#include "SDL_image.h"


#include "lua.h"
#include "lualib.h"
#include "lauxlib.h"


struct Font
{
	GLuint tex;
	int w,h;
};

struct Font load_font(const char *font_path);
void draw(const char *buf, int buflen, struct Font font, int width, float (*palette)[3]);

int main(argc, argv)
{
	lua_State *L;
	int width, height;
	const char *font_path;
	
	SDL_Surface *screen;
	int quit;
	
	SDL_Event event;
	
	struct Font font;
	const char *buffer = "";
	int buflen;
	
	float palette[16][3];
	
	int i,j;
	
	int now, before;
	int count;
	
	/* Init SDL */
	
	if(SDL_Init(SDL_INIT_VIDEO) < 0)
	{
		printf("Couldn't initialize SDL: %s\n", SDL_GetError());
		exit(1);
	}
	
	SDL_EnableUNICODE(1);
	
	/* Init Lua */
	
	L = luaL_newstate();
	luaL_openlibs(L);
	
	if(luaL_dofile(L, "client.conf") != 0)
	{
		printf("Couldn't read client.conf\n");
		exit(1);
	}
	
	lua_getglobal(L,"font");
	font_path = lua_tostring(L,-1);
	
	lua_getglobal(L,"width");
	width = lua_tointeger(L,-1);
	
	lua_getglobal(L,"height");
	height = lua_tointeger(L,-1);
	
	lua_pop(L,3);
	
	printf("Font: %s, %dx%d\n", font_path, width, height);
	
	/* Read palette */
	
	lua_getglobal(L,"palette");
	
	for(i=1 ; i <= 16 ; i++)
	{
		lua_pushnumber(L,i);
		lua_gettable(L,-2);
		
		/* we now how the i-th color table on top of the stack */
		
		for(j=1 ; j<=3 ; j++)
		{
			lua_pushnumber(L,j);
			lua_gettable(L,-2);
			palette[i-1][j-1] = lua_tointeger(L,-1) / 255.0;
			lua_pop(L,1);
		}
		
		lua_pop(L,1);
	}
	
	lua_pop(L,1);
	
	printf("Palette: ");
	for(i=0 ; i < 3*16 ; i++)
		printf("%f ", ((float*) palette)[i]);
	printf("\n");
	
	/* Load actual code */
	
	if(luaL_dofile(L, "client.lua") != 0)
	{
		printf("Couldn't read client.lua\n");
		exit(1);
	}
	
	
	/* Start video */
	
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	screen = SDL_SetVideoMode(width, height, 32, SDL_HWSURFACE|SDL_DOUBLEBUF|SDL_OPENGL);
	
	font = load_font(font_path);
	glBindTexture( GL_TEXTURE_2D, font.tex );
	
	glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
	glViewport( 0, 0, width, height );
	glClear( GL_COLOR_BUFFER_BIT );
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glOrtho(0.0f, width, height, 0.0f, -1.0f, 1.0f);
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
	
	/* Start the connection */
	
	lua_getglobal(L, "init");
	lua_pushnumber(L, (int) width/font.w);
	lua_pushnumber(L, (int) height/font.h);
	lua_call(L,2,1);
	if(!lua_toboolean(L,-1))
	{
		exit(1);
	}
	lua_pop(L,1);
	
	
	/* Loop */
	
	count = 0;
	before = SDL_GetTicks();
	
	quit = 0;
	while(!quit)
	{
		while(SDL_PollEvent(&event))
		{
			switch(event.type)
			{
				case SDL_KEYDOWN:
					lua_getglobal(L, "input");
					lua_pushnumber(L, event.key.keysym.sym);
					lua_pushnumber(L, event.key.keysym.unicode);
					lua_call(L, 2, 0);
					break;
				case SDL_KEYUP:
					lua_getglobal(L, "input");
					lua_pushnumber(L, -(signed int)event.key.keysym.sym);
					lua_call(L, 1, 0);
					break;
				case SDL_QUIT:
					quit = 1;
					break;
				default:
					break;
			}
		}
		
		/* check net connection for new data */
		
		lua_getglobal(L,"update");
		lua_call(L,0,0);
		
		/* draw screen */
		
		lua_getglobal(L,"screen");
		buffer = lua_tostring(L,-1);
		buflen = lua_objlen(L,-1);
		lua_pop(L,1);
				
		draw(buffer, buflen, font, width, palette);
		
		/* check for lost connection */
		
		lua_getglobal(L,"disco");
		if(lua_toboolean(L,-1))
		{
			printf("Connection to server was closed\n");
			quit = 1;	
		}
		lua_pop(L,1);
		
		
		/* count fp */
		
		now = SDL_GetTicks();
		if(now > before + 10000)
		{
			printf("FPS: %f\n", (float)count/(now-before)*1000.0);
			before = now;
			count = 0;
		}
		
		count ++;
	}
	
	/* Clean up */
	glDeleteTextures(1, &(font.tex));
	SDL_Quit();

	return 0;
}


struct Font load_font(const char *font_path)
{
	GLuint texture;
	SDL_Surface *surface;
	GLenum texture_format;
	GLint nOfColors;
	struct Font font;
	
	surface = IMG_Load(font_path);
	
	if(!surface)
	{
		printf("Couldn't load font %s\n", font_path);
		exit(1);
	}
	
	nOfColors = surface->format->BytesPerPixel;
	if (nOfColors == 4)
	{
		if (surface->format->Rmask == 0x000000ff)
			texture_format = GL_RGBA;
		else
			texture_format = GL_BGRA;
	}
	else if (nOfColors == 3)     // no alpha channel
	{
		if (surface->format->Rmask == 0x000000ff)
			texture_format = GL_RGB;
		else
			texture_format = GL_BGR;
	}
	else
	{
		printf("Image %s is not truecolor. Cannot load\n", font_path);
		exit(1);
	}

	/* Have OpenGL generate a texture object handle for us */
	glGenTextures(1, &texture);

	/* Bind the texture object */
	glBindTexture(GL_TEXTURE_2D, texture);
	
	/* Set the texture's stretching properties */
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	
	/* For color modulation */
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	
	/* Edit the texture object's image data using the information SDL_Surface gives us */
	glTexImage2D(GL_TEXTURE_2D, 0, nOfColors, surface->w, surface->h, 0,
		texture_format, GL_UNSIGNED_BYTE, surface->pixels);

	font.tex = texture;
	font.w = surface->w/16;
	font.h = surface->h/16;

	SDL_FreeSurface(surface);
	
	return font;
}

void draw(const char *buf, int buflen, struct Font font, int width, float (*palette)[3])
{
	int nw;
	int x,y;
	int i;
	float cell = 1./16.;
	unsigned char u,v;
	
	float *color;
	
	nw = width/font.w;
	
	glClear(GL_COLOR_BUFFER_BIT);

	/* Background */
	
	glDisable(GL_TEXTURE_2D);

	i = 0;
	x = 0;
	y = 0;
	
	glBegin(GL_QUADS);
		
	while(i < buflen)
	{
		color = palette[(buf[i+1] & 0xF0) >> 4];

		glColor3fv(color);

		glVertex2f(x*font.w, y*font.h);
		glVertex2f((x+1)*font.w, y*font.h);
		glVertex2f((x+1)*font.w, (y+1)*font.h);
		glVertex2f(x*font.w, (y+1)*font.h);
		
		/* Check end of line */
		i = i + 2;
		x++;
		if(x == nw)
		{
			x = 0;
			y++;
		}
	}
	
	glEnd();
	
	
	/* Glyphs */
	
	glEnable(GL_TEXTURE_2D);
	
	i = 0;
	x = 0;
	y = 0;
	
	glBegin(GL_QUADS);
	
	while(i < buflen)
	{

		u = (unsigned char) buf[i] % 16;
		v = (unsigned char) buf[i] / 16;
				
		color = palette[buf[i+1] & 0x0F];
		glColor3fv(color);
		
		glTexCoord2f(u*cell, v*cell);
		glVertex2f(x*font.w, y*font.h);
		
		glTexCoord2f((u+1)*cell, v*cell);
		glVertex2f((x+1)*font.w, y*font.h);
		
		glTexCoord2f((u+1)*cell, (v+1)*cell);
		glVertex2f((x+1)*font.w, (y+1)*font.h);
		
		glTexCoord2f(u*cell, (v+1)*cell);
		glVertex2f(x*font.w, (y+1)*font.h);

		/* Check end of line */
		i = i + 2;
		x++;
		if(x == nw)
		{
			x = 0;
			y++;
		}
	}
	
	glEnd();

	
	SDL_GL_SwapBuffers();
	return;
}
