-- lalonde
-- Copyright (C) 2010 Noé Falzon <firstname.lastname@aliceadsl.fr>

-- This program is  free software:  you can redistribute it and/or modify it
-- under  the terms  of the  GNU General Public License  as published by the
-- Free Software Foundation, version 3 of the License.
--
-- This program  is  distributed  in the hope  that it will  be useful,  but
-- WITHOUT  ANY WARRANTY;  without  even the implied  warranty of MERCHANTA-
-- BILITY  or  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the  GNU General Public License  along
-- with this program. If not, see <http://www.gnu.org/licenses/>.

require "CP437"
require "8859-1"
require "8859-15"

local MF = math.floor
local SC = string.char
local SB = string.byte
local TI = table.insert
local FMT = string.format

-- basic encoding conversion.
-- internally uses Unicode code points on three chars.

function cp_to_string(cp)
	
	local a,b,c
	local cp = cp
	
	c = cp % 256
	cp = MF(cp/256)
	b = cp % 256
	cp = MF(cp/256)
	a = cp % 256
	
	return SC(a,b,c)
end

-- convert the numerical tables to string tables
local tmp = {}

for k,enc in pairs(encodings) do
	
	tmp[k] = {}
		
	for i,v in pairs(enc) do
				
		tmp[k][SC(i)] = cp_to_string(v)
	end
end

encodings = tmp

-- make reverse tables
local reverse_encodings = {}

for k,enc in pairs(encodings) do
	
	local t = {}

	for i,v in pairs(enc) do
		t[v] = i
	end
	
	reverse_encodings[k] = t
end

to_unicode = function(str,enc)

	if enc == "UTF8" then
		return utf8_to_unicode(str)
	end
	
	if not encodings[enc] then
		error("Encoding " .. enc .. " is not available")
	end

	return str:gsub(".", encodings[enc])
end

from_unicode = function(str,enc)

	if enc == "UTF8" then
		return unicode_to_utf8(str)
	end
	
	if not reverse_encodings[enc] then
		error("Encoding " .. enc .. " is not available")
	end

	return str:gsub("...", function(c)
		local c2 = reverse_encodings[enc][c]
		
		if not c2 then
			c2 = "?"
		end
		
		return c2
	end)
end

-- utf8 stuff, modeled on LPeg example

-- decode a two-byte UTF-8 sequence
local function f2 (s)
	local c1, c2 = SB(s, 1, 2)
	return c1 * 64 + c2 - 12416
end

-- decode a three-byte UTF-8 sequence
local function f3 (s)
	local c1, c2, c3 = SB(s, 1, 3)
	return (c1 * 64 + c2) * 64 + c3 - 925824
end

-- decode a four-byte UTF-8 sequence
local function f4 (s)
	local c1, c2, c3, c4 = SB(s, 1, 4)
	return ((c1 * 64 + c2) * 64 + c3) * 64 + c4 - 63447168
end

local function try_utf8_char(s,pos)
	
	-- 1 byte
	local _,_,c = s:find("^([%x\1-\127])", pos)
	
	if c then
		return c	
	end
	
	-- 2 bytes
	_,_,c = s:find("^([\194-\223][\128-\191])", pos)
	
	if c then
		return c	
	end
	
	-- 3 bytes
	_,_,c = s:find("^([\224-\239][\128-\191][\128-\191])", pos)
	
	if c then
		return c	
	end
	
	-- 4 bytes
	_,_,c = s:find("^([\240-\244][\128-\191][\128-\191][\128-\191])", pos)
	
	if c then
		return c	
	end
	
end

function utf8_to_unicode(s)
	local i = 1
	local cp = {}

	while i <= #s do
		
		local c = try_utf8_char(s,i)
		
		if #c == 1 then
			TI(cp, SB(c))
			i = i+1
		elseif #c == 2 then
			TI(cp, f2(c))
			i = i+2
		elseif #c == 3 then
			TI(cp, f3(c))
			i = i+3
		elseif #c == 4 then
			TI(cp, f4(c))
			i = i+4
		else
			-- malformed utf8 sequence
			error("malformed utf8 sequence")
		end
		
	end
	
	for i,v in ipairs(cp) do
		cp[i] = cp_to_string(v)
	end
	
	return table.concat(cp)
end

function unicode_to_utf8(str)
	
	return str:gsub("...", function(s)
	
		local a,b,c = s:byte(1,3)
		
		local cp = (a*256+b)*256+c
	
		if cp <= 0x7F then
			return SC(c)		
		
		elseif cp <= 0x7FF then
		
			return SC (MF(cp/64) + 192, cp%64 + 128)
		
		elseif cp <=0xFFFF then
			local x,y,z
			
			z = cp % 64 + 128
			cp = MF(cp/64)
			y = cp % 64 + 128
			x = MF(cp/64) + 224
			
			return SC(x,y,z)
		else
			local w,x,y,z
			
			z = cp % 64 + 128
			cp = MF(cp/64)
			y = cp % 64 + 128
			cp = MF(cp/64)
			x = cp % 64 + 128
			w = MF(cp/64) + 240
			
			return SC(w,x,y,z)
		end
		
	end)
end

function cp_to_utf8(cp)
	
	local s = cp_to_string(cp)
	return unicode_to_utf8(s)
end

function reencode(s,src,dst)
	
	local uni = to_unicode(s,src)
	return from_unicode(uni,dst)
end
