-- lalonde
-- Copyright (C) 2010 Noé Falzon <firstname.lastname@aliceadsl.fr>

-- This program is  free software:  you can redistribute it and/or modify it
-- under  the terms  of the  GNU General Public License  as published by the
-- Free Software Foundation, version 3 of the License.
--
-- This program  is  distributed  in the hope  that it will  be useful,  but
-- WITHOUT  ANY WARRANTY;  without  even the implied  warranty of MERCHANTA-
-- BILITY  or  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the  GNU General Public License  along
-- with this program. If not, see <http://www.gnu.org/licenses/>.

require "class"

require "Lobby"
require "keys"

Client = class()

function Client:init(sock)
	
	self.sock = sock
	self.name = self.sock:getpeername()
	
	-- terminal
	self.w = 0
	self.h = 0
	self.term = nil
	self.term2 = nil
	
	
	-- current view, starts in Lobby
	self.view = Lobby(self)
	
	return self
end

function Client:input(data)
	
	-- Check for special control sequences
	
	-- requested terminal size
	local w,h = data:match("tsize (%d+) (%d+)")
	w = tonumber(w)
	h = tonumber(h)
	if w and h then
		self.w = w
		self.h = h
		
		self:resizeterm(1)
		self:resizeterm(2)
		
		print(self.name .. " sets terminal size to " .. w .. "," .. h)
		return
	end
	
	-- nick change
	local nick = data:match("nick (%w+)")
	if nick then
		
		print(self.name .. " is now known as " .. nick)
		self.name = nick
		
		return
	end
	
	-- the rest is interpreted as a key press and transmitted to current view
	-- as a utf8 string of the unicode char, or the SDL name of the key
	
	-- simple key press/up
	if data:sub(1,1) == "k" then
		
		local key = tonumber(data:sub(2))
		
		local dir = 'down'
		if key < 0 then
			dir = 'up'
			key = -key
		end
		
		self.view:input(keys[key],dir)
	
	-- unicode code point
	else
		local cp = tonumber(data)
		self.view:input(cp_to_utf8(cp),'unicode')
	end
end

-- called at disconnection
function Client:cleanup()
	
	self.view:cleanup()
end

function Client:resizeterm(n)
	
	local t = {}
	
	for l = 1,self.h do
		t[l] = {}
		for c = 1,self.w do
			t[l][c] = " \15"
		end
	end
	
	if n == 1 then	
		self.term = t
	else
		self.term2 = t
	end
end

function Client:termsequal()
	
	for l = 1,self.h do
		for c = 1,self.w do
			
			if self.term[l][c] ~= self.term2[l][c] then
				return false
			end
		end
	end
	
	return true
end

function Client:sendterm()

	if not self.term then
		return
	end
	
	local t = {}
	
	-- concat each line
	for i,l in ipairs(self.term) do
		t[i] = TC(l)
	end
	
	-- and all lines together
	t = TC(t)
	
	-- send to client
	Lalonde.Net:send(t,self.sock)	
end

function Client:update()
	
	if not self.term then
		return
	end
	
	local disco = self.view:update()
	
	-- read stuff from current view
	-- put in the second buffer

	self.view:draw(self.term2)
	
	-- compare with previous buffer, and send if different
	if not self:termsequal() then
	
		-- swap terms
		self.term,self.term2 = self.term2,self.term
		
		self:sendterm()
	end
	
	if disco then
		return true
	end
end
