-- lalonde
-- Copyright (C) 2010 Noé Falzon <firstname.lastname@aliceadsl.fr>

-- This program is  free software:  you can redistribute it and/or modify it
-- under  the terms  of the  GNU General Public License  as published by the
-- Free Software Foundation, version 3 of the License.
--
-- This program  is  distributed  in the hope  that it will  be useful,  but
-- WITHOUT  ANY WARRANTY;  without  even the implied  warranty of MERCHANTA-
-- BILITY  or  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the  GNU General Public License  along
-- with this program. If not, see <http://www.gnu.org/licenses/>.

-- "graphic" utilities

module(...,package.seeall)

char = function(t,l,c,char,col,bcol)
	
	if l > #t or c > #t[1] then
		return
	end
	
	local s = t[l][c]
	
	local char = char and char:byte(1) or s:byte(1)
	local bcol = bcol or MF(s:byte(2) / 16)
	local col = col or s:byte(2) % 16
	
	t[l][c] = SC(char, bcol*16 + col)
end

rect = function(t,l,c,h,w,char_,col,bcol)
	
	for i = l,l+h-1 do
		for j = c,c+w-1 do
			char(t,i,j,char_,col,bcol)
		end
	end
end

hline = function(t,l,c,w,char_,col,bcol)
	
	for j = c,c+w-1 do
		char(t,l,j,char_,col,bcol)
	end
end

vline = function(t,l,c,h,char_,col,bcol)
	
	for i = l,l+h-1 do
		char(t,i,c,char_,col,bcol)
	end
end

rectline = function(t,l,c,h,w,char_,col,bcol)
	
	hline(t,l,c,w,char_,col,bcol)
	hline(t,l+h-1,c,w,char_,col,bcol)
	vline(t,l+1,c,h-2,char_,col,bcol)
	vline(t,l+1,c+w-1,h-2,char_,col,bcol)
end

text = function(t,l,c,str,col,bcol)
	
	for i = 1,#str do
		char(t,l,c+i-1,str:sub(i,i),col,bcol)
	end	
end

centertext = function(t,l,str,col,bcol)
	
	text(t,l,MF((#t[1] - #str)/2) + 1,str,col,bcol)
end

textbox = function(t,l,c,h,w,str,col,bcol)
	
	-- fill the background
	rect(t,l,c,h,w,' ',col,bcol)
	
	-- write text	
	local i,j = l,c
	
	for word in str:gmatch("%S+") do
		
		-- newline if necessary
		if j + #word > c + w then
			i = i + 1
			j = c
		end
		
		-- stop if too low
		if i >= l + h then
			return
		end
		
		-- print word	
		text(t,i,j,word,col,bcol)
		
		-- advance counter + space
		j = j + #word + 1
	
	end	
end

-- change the text color intext by using '!abbrev!'
ftext = function(t,l,c,str,bcol)
	
	local col
	local pos = c
	
	local i = 1
	
	while i <= #str do
		local ch = str:sub(i,i)
		
		if ch == '!' then
			local a,b,capt = str:find("!(.-)!", i)
			
			-- change color
			col = abbr_palette[capt]

			-- skip the control sequence
			i = b + 1
		else
			
			char(t,l,pos,str:sub(i,i),col,bcol)
			pos = pos + 1
			i = i + 1
		end
	end	
end
