-- lalonde
-- Copyright (C) 2010 Noé Falzon <firstname.lastname@aliceadsl.fr>

-- This program is  free software:  you can redistribute it and/or modify it
-- under  the terms  of the  GNU General Public License  as published by the
-- Free Software Foundation, version 3 of the License.
--
-- This program  is  distributed  in the hope  that it will  be useful,  but
-- WITHOUT  ANY WARRANTY;  without  even the implied  warranty of MERCHANTA-
-- BILITY  or  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
-- License for more details.
--
-- You should have received a copy of the  GNU General Public License  along
-- with this program. If not, see <http://www.gnu.org/licenses/>.

package.path = package.path .. ";../common/?.lua;../encoding/?.lua;../game/?.lua"

require "utils"
require "palette"
require "TcpServer"
require "ClientManager"
require "encoding"

-- Singleton "class" Lalonde

Lalonde = {}

--[[ Initialization ]]

function Lalonde:init()

	-- TcpServer

	local port = arg[1] or 6666	-- take first CLI argument as the port

	self.Net = TcpServer(port,0,0)

	if not self.Net then
		print("Could not run server on port " .. port .. ". Exiting.")
		return
	end

	print("Server up", self.Net.ip .. ":" .. self.Net.port)

	-- Clients

	self.Clients = ClientManager()

	-- Callbacks

	self.Net.cbconnect = function(sock)

		self.Clients:addclient(sock)
	end

	self.Net.cbrecv = function(data,sock)
		
		self.Clients:input(data,sock)
	end

	self.Net.cbdisconnect = function(sock)
		
		self.Clients:removeclient(sock)
	end

end

--[[Main loop]]

function Lalonde:main()

	self.quit = false

	local old = os.time()
	local count = 0

	while not self.quit do

		-- check new clients and incoming input
		self.Net:update()
		
		-- update world
		if self.GameInstance then
			self.GameInstance:update()
		end
		
		-- update clients
		self.Clients:updateall()
		
		-- fps
		local now = os.time()
		if now > old + 10 then
			local fps = count/(now-old)
			print("FPS: ", fps)
			old = now
			count = 0
		end
		
		count = count + 1
	end

end

function Lalonde:cleanup()

	-- Clean up
	print "Closing down connections"
	self.Net:close()
	print "Exiting server"

end

-- Run it!

Lalonde:init()
Lalonde:main()
Lalonde:cleanup()
